/***************************************************************************/
/***                                                                     ***/
/***                   solve.c   (Version 2.0)                           ***/
/***                                                                     ***/
/***       Jochen Renz, Ronny Fehling, Bernhard Nebel  - July 1999       ***/
/***                                                                     ***/
/***         fehling, nebel, renz@informatik.uni-freiburg.de             ***/
/***                                                                     ***/
/***          http://www.informatik.uni-freiburg.de/~sppraum             ***/
/***                                                                     ***/
/***                      Institut fuer Informatik                       ***/
/***                     Albert-Ludwigs-Universitaet                     ***/
/***                           Am Flughafen 17                           ***/
/***                       79110 Freiburg, Germany                       ***/
/***                                                                     ***/
/***************************************************************************/

/*
solve [-bdefnpqrsv] [-w[v]] [-z[p]] [-m[f] number] [-g number] [-ctx file] [-o file [number]] [-h[a|A file] order] [-li <file>] [-le <file>] [file1 file2 ...]  
    solves RCC8 CSPs.
  solve [-bdefnpqrsvwz] [-m number] [-ctx file] [-o file [number]] [-h[a|A file] order] [file1 file2 ...]  
       solves RCC8 CSPs.

    options are:
    -b    return brief summary info on stdout (even if -q)
    -c <file> use the coverage info in file to split relations
    -d    give debugging info on stderr
    -e    use environment constrainedness to select next node (van Beek)
    -f    fixed ordering of variables 
    -g <number>  solve only instances larger than <number> 
    -h[a|A <file>]  <order>  use the orthogonal combination of the different 
                    heuristic methods. (a = always run all methods, the maximal
                    number of visited nodes of following methods is restricted
                    by the number required by the previously best method; 
                    A = run all methods and write output in <file>) 
                    <order> specifies the order of the methods. 
                    1=H^/dyn/loc, 2=H^/dyn/glo/, 3=H^/sta/loc, 4=H^/sta/glo, 
                    A=B^/dyn/loc, B=B^/dyn/glo/, C=B^/sta/loc, D=B^/sta/glo,
                    a=B/dyn/loc, b=B/dyn/glo/, c=B/sta/loc, d=B/sta/glo,
                    s=C/dyn/loc, t=C/dyn/glo/, u=C/sta/loc, v=C/sta/glo, 
		    S=Q/dyn/loc, T=Q/dyn/glo/, U=Q/sta/loc, V=Q/sta/glo,
                    i=I/dyn/loc, j=I/dyn/glo/, k=I/sta/loc, l=I/sta/glo,
                    <order>=2B4a means that the order is 2,B,4 and finally a.
                    NOTE: It is assumed that the files hornsplit, 
       	            closebasesplit, c8split, q8split and intersplit exist. 
                    This option overrides the options -c <file>, -f, and -e.  
    -li <file>   Include all headers of the logfile <file>
    -le <file>   Exclude all headers of the logfile <file>
                 NOTE: `-li <file>' must be given before `-le <file>'. 
    -m[f] <number> maximal number of nodes to visit (default NV_UPPER_LIMIT)
                   if -mf is set (f=factor), then the maximal number of nodes 
                   to visit is the instance size multiplicated by <number> 
    -n    restrict summary to negative (inconsistent) cases
    -o <file> <number> solve only those CSP that are tagged with [-1,0,1] 
                       (default -1) (should be a log-file)
    -p    restrict summary to positive (consistent) cases
    -q    return 0 if (last) CSP was consistent (suppresses -v) 
    -r    use random variable ordering 
    -s    return summary info on stdout (even if -q)
    -t <file> print return summary to file after typeid changes
    -v    give verbose output on stderr after solving a CSP
    -w[v] use weighted queue scheme for computing path-consistency 
          (v = van Beek)
    -x <file> print CSPs that could not been solved to file
    -z[p] compute path-consistency only (p=print path-consistent CSP to stdout)

    file[i] are CSP-input files. If none specified, solve uses stdin. 
    More than one CSP can be specified in each file (each must be 
    terminated by a point '.').

    Format of the CSP should be <node1> <node2> '(' <relations> ')'.
    First line of CSP should contain maximum node id and typeid (string).";
*/


const char *usagestring = "\
usage: solve [-bdefnpqrsv] [-w[v]] [-z[p]] [-m[f] number] [-g number] [-ctx file] [-o file [number]] [-h[a|A file] order] [-li <file>] [-le <file>] [file1 file2 ...]  
    solves RCC8 CSPs.

    options are:
    -b    return brief summary info on stdout (even if -q)
    -c <file> use the coverage info in file to split relations
    -d    give debugging info on stderr
    -e    use environment constrainedness to select next node (van Beek)
    -f    fixed ordering of variables 
    -g <number>  solve only instances larger than <number> 
    -h[a|A <file>]  <order>  use the orthogonal combination of the different 
                    heuristic methods. (a = always run all methods, the maximal
                    number of visited nodes of following methods is restricted
                    by the number required by the previously best method; 
                    A = run all methods and write output in <file>) 
                    <order> specifies the order of the methods. 
                    1=H^/dyn/loc, 2=H^/dyn/glo/, 3=H^/sta/loc, 4=H^/sta/glo, 
                    A=B^/dyn/loc, B=B^/dyn/glo/, C=B^/sta/loc, D=B^/sta/glo,
                    a=B/dyn/loc, b=B/dyn/glo/, c=B/sta/loc, d=B/sta/glo,
                    s=C/dyn/loc, t=C/dyn/glo/, u=C/sta/loc, v=C/sta/glo, 
		    S=Q/dyn/loc, T=Q/dyn/glo/, U=Q/sta/loc, V=Q/sta/glo,
                    i=I/dyn/loc, j=I/dyn/glo/, k=I/sta/loc, l=I/sta/glo,
                    <order>=2B4a means that the order is 2,B,4 and finally a.
                    NOTE: It is assumed that the files hornsplit, 
       	            closebasesplit, c8split, q8split and intersplit exist. 
                    This option overrides the options -c <file>, -f, and -e.  
    -li <file>   Include all headers of the logfile <file>
    -le <file>   Exclude all headers of the logfile <file> 
                 NOTE: `-li <file>' must be given before `-le <file>'. 
    -m[f] <number> maximal number of nodes to visit (default NV_UPPER_LIMIT)
                   if -mf is set (f=factor), then the maximal number of nodes 
                   to visit is the instance size multiplicated by <number> 
    -n    restrict summary to negative (inconsistent) cases
    -o <file> <number> solve only those CSP that are tagged with [-1,0,1] 
                       (default -1) (should be a log-file)
    -p    restrict summary to positive (consistent) cases
    -q    return 0 if (last) CSP was consistent (suppresses -v) 
    -r    use random variable ordering 
    -s    return summary info on stdout (even if -q)
    -t <file> print return summary to file after typeid changes
    -v    give verbose output on stderr after solving a CSP
    -w[v] use weighted queue scheme for computing path-consistency 
          (v = van Beek)
    -x <file> print CSPs that could not been solved to file
    -z[p] compute path-consistency only (p=print path-consistent CSP to stdout)

    file[i] are CSP-input files. If none specified, solve uses stdin. 
    More than one CSP can be specified in each file (each must be 
    terminated by a point '.').

    Format of the CSP should be <node1> <node2> '(' <relations> ')'.
    First line of CSP should contain maximum node id and typeid (string).";



#define HARD_NODES 	10000


#include "rcc8.h"
#include "rcc8io.h"
#include "rcc8op.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/time.h>

/* should be in stdlib.h ! */
void srand48(long seedval);

/* should be in resource.h ! */
 int getrusage(int who, struct rusage *rusage);


/* switches */
int swhard;
int swdebug;
int swverbose;
int swquiet;
int swsummary;
int swbrief;
int swpositive;
int swnegative;
int swposneg;
int swsplitinfo;
int swrandomorder;
int swwqueue;
int swenvcons;
int swfixed;
int swsolveonly;
int swonlypathcons;
int swpcprint;
int swtesting;
int swmaxvisit;
int swfactor;
int sworthogonal;
int swallorthogonal; 
int swheuristicout;
int swheader;
int swminsize;

/* statistical info */
int taggedwith =5;
int maxnodeid;
char csptype[100];
char lasttype[100];
int consistent;
int cntconsistent;
int pathconsistent;
int cntpathconsistent;
int pcops;
int pcits;
int nodesvisited;
int searchdepth;
int maxdepth;
double cputicks;
double cpumicro;
int trial;
int rectrial;
long int tpcops[MAXTRIAL];
long int tpcits[MAXTRIAL];
long int tnodesvisited[MAXTRIAL];
int tmaxdepth[MAXTRIAL];
double tcputicks[MAXTRIAL];
int tconsistent[MAXTRIAL];
double cumcputicks;
double cumcpumicro;
int cntsize;
int size;
int maxnodes=NV_UPPER_LIMIT;
int minsize=0;
 
struct header headers[MAXHEADER]; 
/* int maxheader;  */

RELTYPE splitinfo[MAXSET][9];
RELTYPE horninfo[MAXSET][9];
RELTYPE c8info[MAXSET][9];
RELTYPE q8info[MAXSET][9];
RELTYPE closebaseinfo[MAXSET][9];
RELTYPE intersectinfo[MAXSET][9];

#if defined(DYNAMIC)
RELTYPE **csp;
struct entry **entry;
char **mark;
int **consval;
int prevmax=0;
#else 
RELTYPE csp[MAXCSP][MAXCSP];
struct entry entry[MAXWT][MAXCSP*MAXCSP];
char mark[MAXCSP][MAXCSP];
int consval[MAXCSP][MAXCSP];
#endif


/* files */
FILE *infile;
char infilename[200];
FILE *splitfile;
char splitfilename[200];
FILE *solveonlyfile;
char solveonlyname[200];
FILE *summaryfile;
char summaryname[200];
FILE *hardfile;
char hardfilename[200];
FILE *heuristicfile;
char heuristicfilename[200];
FILE *headerfile;
char headerfilename[200];

/* timing info */
struct tms local_before, local_after, global_before, global_after;
struct rusage rus_glb_before,rus_glb_after,rus_lcl_before,rus_lcl_after;

/* random variable */
unsigned int randomseed = 0;



void initrand(unsigned int seed)
{
  if (seed == 0)
    randomseed = time(NULL);
  else
    randomseed = seed;

  srand48(randomseed);
}



int path_cons(csp, maxnodeid, node1, node2)
     int       maxnodeid, node1, node2;
#if defined(DYNAMIC) 
     RELTYPE   **csp;
#else 
     RELTYPE csp[MAXCSP][MAXCSP];
#endif
{
  if (swwqueue==1) 
    return(path_cons_wq(csp,maxnodeid,node1,node2));
  else if(swwqueue==2)
    return(path_cons_vb(csp,maxnodeid,node1,node2));
  else
    return(path_cons1(csp,maxnodeid,node1,node2));
}



void usage()
{
  fprintf(stderr,"%s\n",usagestring);
  exit(-1);
}



int int_compare(const void *p1, const void *p2){
  const long int *i1=p1, *i2=p2;
  return(((*i1 < *i2) ? -1 
	  : ((*i1 == *i2) ? 0 : 1)));
}



int double_compare( const void *p1, const void *p2){
  const double *i1=p1, *i2=p2;
  return(((*i1 < *i2) ? -1 
	  : ((*i1 == *i2) ? 0 : 1)));
}



void init_statistics()
{
  trial = 0;
  rectrial = 0;
  cntpathconsistent = 0;
  cntconsistent = 0;
  cntsize = 0; 
}

void heuristic_summary(int minnodes, char *order, char first, int *ordernode, FILE *file)
{
  int i=0;
  char name[20];
  fprintf(file,"%s cons: %i nodes: %i method: %c\n",csptype,consistent,minnodes,first); 
  while(order[i]!='\0') {
    if(ordernode[i]==0) {
      i++;
      continue;
    }
    switch(order[i]) {
    case '1': strcpy(name,"H^/dyn/loc");break;
    case '2': strcpy(name,"H^/dyn/glo");break;
    case '3': strcpy(name,"H^/sta/loc");break;
    case '4': strcpy(name,"H^/sta/glo");break;
    case 'A': strcpy(name,"B^/dyn/loc");break;
    case 'B': strcpy(name,"B^/dyn/glo");break;
    case 'C': strcpy(name,"B^/sta/loc");break;
    case 'D': strcpy(name,"B^/sta/glo");break;
    case 'a': strcpy(name,"B/dyn/loc");break;
    case 'b': strcpy(name,"B/dyn/glo");break;
    case 'c': strcpy(name,"B/sta/loc");break;
    case 'd': strcpy(name,"B/sta/glo");break;
    case 'S': strcpy(name,"Q/dyn/loc");break;
    case 'T': strcpy(name,"Q/dyn/glo");break;
    case 'U': strcpy(name,"Q/sta/loc");break;
    case 'V': strcpy(name,"Q/sta/glo");break;
    case 's': strcpy(name,"C/dyn/loc");break;
    case 't': strcpy(name,"C/dyn/glo");break;
    case 'u': strcpy(name,"C/sta/loc");break;
    case 'v': strcpy(name,"C/sta/glo");break;
    case 'i': strcpy(name,"I/dyn/loc");break;
    case 'j': strcpy(name,"I/dyn/glo");break;
    case 'k': strcpy(name,"I/sta/loc");break;
    case 'l': strcpy(name,"I/sta/glo");break;
    default: strcpy(name,"unknown");
    }
    fprintf(file,"%s: %i\n",name,ordernode[i]);
    i++;
  }
  fflush(file);
}

void print_summary(int brief, char *lasttype, FILE *file)
{
  float sum, avg, med, seventy, ninety, ninetynine, hundred;
  register int i;

  sum = 0.0;
  for (i = 0; i < trial; i++)
    sum += (float)(tcputicks[i]);

  if(!swtesting && !brief) {
    fprintf(file,"time used:\n");
    fprintf(file,"global: %f sec\n",cumcputicks);
    fprintf(file,"local:  %f sec\n",sum);
    fprintf(file,"number: %d\n",trial);
  }

  if (!brief) fprintf(file,"SUMMARY INFORMATION:\n");
  if (trial < 1) {
    fprintf(file,"  No CSP given.\n");
    return;
  }

  if (brief) fprintf(file,"%s \n",lasttype);
  else 
    fprintf(file,"  CSP size=%5.1f (%s), last type=%s\n",
	   ((float)cntsize)/((float)(trial)),
	   ((size == 0) ? "avg." : "all"),
	   lasttype);

  if(swonlypathcons)
    fprintf(file,"  Only checked path-consistency\n");
  if (brief){
      fprintf(file," %5.2f (path) %5.2f (global) ",
		    (((float)cntpathconsistent)/((float)trial))*100.0,
		    (((float)cntconsistent)/((float)trial))*100.0);
  }
  else{
      fprintf(file,"  Consistent: path=%6.2f%%, global=%6.2f%%\n",
	      (((float)cntpathconsistent)/((float)trial))*100.0,
	      (((float)cntconsistent)/((float)trial))*100.0);
  }
  if (swpositive) {
    if (brief) fprintf(file,"C \n");
    else fprintf(file,"  Statistics on CONSISTENT cases only:\n");
  }
  else if (swnegative) {
    if (brief) fprintf(file,"I \n");
    else fprintf(file,"  Statistics on INCONSISTENT cases only:\n");
  }
  else if (brief) fprintf(file,"\n");

  trial = rectrial;

  if (trial <= 0) {
    if (brief) fprintf(file," -------\n");
    else fprintf(file,"  *** no cases\n");
    return;
  }

  if(!brief)
    fprintf(file,"\n\t    average  percentile:   (50)\t   (70)\t   (90)\t   (99)\t  (100)\n");

  avg = sum/((float)trial);
  qsort(tcputicks,trial,sizeof(double),&double_compare);
  if (trial == 1) {
    med = (float)tcputicks[0];
    seventy = (float)tcputicks[0];
    ninety = (float)tcputicks[0];
    ninetynine = (float)tcputicks[0];
    hundred = (float)tcputicks[0];
  } else {
    med = ((float)(tcputicks[(trial/2)-1]+tcputicks[trial/2]))/2.0;
    seventy = ((float)(tcputicks[((trial*7)/10)-1]+
		       tcputicks[((trial*7)/10)]))/2.0;
    ninety = ((float)(tcputicks[((trial*9)/10)-1]+
		       tcputicks[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tcputicks[((trial*99)/100)-1]+
		       tcputicks[((trial*99)/100)]))/2.0;
    hundred = (float) tcputicks[trial-1];
  }
  if (brief)
    fprintf(file," %6.5f %6.5f %6.5f %6.5f %6.5f %6.5f (cpu)\n",avg,med,seventy,ninety,ninetynine, hundred);
  else
    fprintf(file,"  CPU time: %6.5f\t\t%6.5f %6.5f %6.5f %6.5f %6.5f\n",
	   avg, med, seventy, ninety, ninetynine, hundred);

  if (!brief) {
    avg = 0.0;
    for (i = 0; i < trial; i++)
      avg += (float)(tpcops[i]);
    avg = avg/((float)trial);
    qsort(tpcops,trial,sizeof(long int),&int_compare);
    if (trial == 1) {
      med = (float)tpcops[0];
      seventy = (float)tpcops[0];
      ninety = (float)tpcops[0];
      ninetynine = (float)tpcops[0];
      hundred = (float)tpcops[0];
    } else {
      med = ((float)(tpcops[(trial/2)-1]+tpcops[trial/2]))/2.0;
      seventy = ((float)(tpcops[((trial*7)/10)-1]+
			 tpcops[((trial*7)/10)]))/2.0;
      ninety = ((float)(tpcops[((trial*9)/10)-1]+
			tpcops[((trial*9)/10)]))/2.0;
      ninetynine = ((float)(tpcops[((trial*99)/100)-1]+
			    tpcops[((trial*99)/100)]))/2.0;
      hundred = (float) tpcops[trial-1];
    }
      fprintf(file,"  PC ops:   %7.0f\t\t%7.0f %7.0f %7.0f %7.0f %7.0f\n",
	      avg, med, seventy, ninety, ninetynine, hundred);
 

      avg = 0.0;
      for (i = 0; i < trial; i++)
	avg += (float)(tpcits[i]);
      avg = avg/((float)trial);
      qsort(tpcits,trial,sizeof(long int),&int_compare);
      if (trial == 1) {
	med = (float)tpcits[0];
	seventy = (float)tpcits[0];
	ninety = (float)tpcits[0];
	ninetynine = (float)tpcits[0];
	hundred = (float)tpcits[0];
      } else {
	med = ((float)(tpcits[(trial/2)-1]+tpcits[trial/2]))/2.0;
	seventy = ((float)(tpcits[((trial*7)/10)-1]+
			   tpcits[((trial*7)/10)]))/2.0;
	ninety = ((float)(tpcits[((trial*9)/10)-1]+
			  tpcits[((trial*9)/10)]))/2.0;
	ninetynine = ((float)(tpcits[((trial*99)/100)-1]+
			      tpcits[((trial*99)/100)]))/2.0; 
	hundred = (float) tpcits[trial-1];
      }
      fprintf(file,"  PC its:   %7.1f\t\t%7.1f %7.1f %7.1f %7.1f %7.1f\n",
	      avg, med, seventy, ninety, ninetynine, hundred);
  }

  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tnodesvisited[i]);
  avg = avg/((float)trial);
  qsort(tnodesvisited,trial,sizeof(long int),&int_compare);
  
  if (trial == 1) {
    med = (float)tnodesvisited[0];
    seventy = (float)tnodesvisited[0];
    ninety = (float)tnodesvisited[0];
    ninetynine = (float)tnodesvisited[0];
    hundred = (float)tnodesvisited[0];
  } else {
    med = ((float)(tnodesvisited[(trial/2)-1]+tnodesvisited[trial/2]))/2.0;
    seventy = ((float)(tnodesvisited[((trial*7)/10)-1]+
		       tnodesvisited[((trial*7)/10)]))/2.0;
    ninety = ((float)(tnodesvisited[((trial*9)/10)-1]+
		       tnodesvisited[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tnodesvisited[((trial*99)/100)-1]+
		       tnodesvisited[((trial*99)/100)]))/2.0;
    hundred = (float) tnodesvisited[trial-1];
  }
  if (brief)
    fprintf(file," %6.5f %6.5f %6.5f %6.5f %6.5f %6.5f (nv)\n",
	   avg, med, seventy, ninety, ninetynine, hundred);
  else
    fprintf(file,"  Nodes v.: %7.1f\t\t%7.1f %7.1f %7.1f %7.1f %7.1f\n",
	   avg, med, seventy, ninety, ninetynine, hundred);


  avg = 0.0;
  for (i = 0; i < trial; i++)
    avg += (float)(tmaxdepth[i]);
  avg = avg/((float)trial);
  qsort(tmaxdepth,trial,sizeof(long int),&int_compare);
  
  if (trial == 1) {
    med = (float)tmaxdepth[0];
    seventy = (float)tmaxdepth[0];
    ninety = (float)tmaxdepth[0];
    ninetynine = (float)tmaxdepth[0];
    hundred = (float)tmaxdepth[0];
  } else {
    med = ((float)(tmaxdepth[(trial/2)-1]+tmaxdepth[trial/2]))/2.0;
    seventy = ((float)(tmaxdepth[((trial*7)/10)-1]+
		       tmaxdepth[((trial*7)/10)]))/2.0;
    ninety = ((float)(tmaxdepth[((trial*9)/10)-1]+
		       tmaxdepth[((trial*9)/10)]))/2.0;
    ninetynine = ((float)(tmaxdepth[((trial*99)/100)-1]+
		       tmaxdepth[((trial*99)/100)]))/2.0;
    hundred = (float) tmaxdepth[trial-1];
  }
  if (brief)
    fprintf(file," %6.5f %6.5f %6.5f %6.5f %6.5f %6.5f (md)\n",
	   avg, med, seventy, ninety, ninetynine, hundred);
  else
    fprintf(file,"  Depth:    %7.1f\t\t%7.1f %7.1f %7.1f %7.1f %7.1f\n",
	   avg, med, seventy, ninety, ninetynine, hundred);
 

  init_statistics();

}

void copysplitfile(int newsplit)
{
  int i,j;

  if(newsplit==1) {
    for (i=0;i<MAXSET;i++) 
      for (j=0; j<9; j++) {
	splitinfo[i][j] = horninfo[i][j];
	}
      }
  if(newsplit==2) {
    for (i=0;i<MAXSET;i++) 
      for (j=0; j<9; j++) {
	splitinfo[i][j] = closebaseinfo[i][j];
	}
      }
  if(newsplit==3) {
    for (i=0;i<MAXSET;i++) 
      for (j=0; j<9; j++) {
	splitinfo[i][j] = c8info[i][j];
	}
      }
 if(newsplit==4) {
    for (i=0;i<MAXSET;i++) 
      for (j=0; j<9; j++) {
	splitinfo[i][j] = q8info[i][j];
	}
      }
 if(newsplit==5) {
    for (i=0;i<MAXSET;i++) 
      for (j=0; j<9; j++) {
	splitinfo[i][j] = intersectinfo[i][j];
	}
      }
}


void readsplitfile(char *s)
{
  int i,j,inp;

  if ((splitfile = fopen(s,"r")) == NULL) {
    fprintf(stderr, "***Relation-cover file '%s' not found\n",s);
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "***Format error in relation-cover file '%s'\n",s);
	exit(-1);
      } else {
	splitinfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "***No eof in relation-cover file '%s'\n",s);
      	exit(-1);
      }
}

void standardsplitfile(void)
{
  int i,j,inp;

  if ((splitfile = fopen("hornsplit","r")) == NULL) {
    fprintf(stderr, "****Relation-cover file 'hornsplit' not found\n");
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "****Format error in relation-cover file 'hornsplit'\n");
	exit(-1);
      } else {
	horninfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "****No eof in relation-cover file 'hornsplit'\n");
      	exit(-1);
      }
  fclose(splitfile);

 if ((splitfile = fopen("c8split","r")) == NULL) {
    fprintf(stderr, "****Relation-cover file 'c8split' not found\n");
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "****Format error in relation-cover file 'c8split'\n");
	exit(-1);
      } else {
	c8info[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "****No eof in relation-cover file 'c8split'\n");
      	exit(-1);
      }
  fclose(splitfile);

 if ((splitfile = fopen("q8split","r")) == NULL) {
    fprintf(stderr, "****Relation-cover file 'q8split' not found\n");
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "****Format error in relation-cover file 'q8split'\n");
	exit(-1);
      } else {
	q8info[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "****No eof in relation-cover file 'q8split'\n");
      	exit(-1);
      }
  fclose(splitfile);


  if ((splitfile = fopen("closebasesplit","r")) == NULL) {
    fprintf(stderr, "****Relation-cover file 'closebasesplit' not found\n");
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "****Format error in relation-cover file 'closebasesplit'\n");
	exit(-1);
      } else {
	closebaseinfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "****No eof in relation-cover file 'closebasesplit'\n");
      	exit(-1);
      }
  fclose(splitfile);

 if ((splitfile = fopen("intersplit","r")) == NULL) {
    fprintf(stderr, "****Relation-cover file 'intersectsplit' not found\n");
    exit(-1);
  }
  fflush(stderr);
  for (i=0;i<MAXSET;i++) 
    for (j=0; j<9; j++) {
      if (fscanf(splitfile,"%d",&inp) != 1) {
	fprintf(stderr, "****Format error in relation-cover file 'intersplit'\n");
	exit(-1);
      } else {
	intersectinfo[i][j] = inp;
      }
    }
  if ((fscanf(splitfile,"%d",&inp) == 1) || (!feof(splitfile))) {
	fprintf(stderr, "****No eof in relation-cover file 'intersplit'\n");
      	exit(-1);
      }
  fclose(splitfile);
}



void getnextcsp(char csp[100])
{
  int d;
  char line[256];

  d = 1;
  while ((d != 0) && (!feof(solveonlyfile))) {
    fgets(line,255,solveonlyfile);
    if (sscanf(line,"%s %d",csp,&d) != 2) d = 1;
    if ((d == taggedwith) && (csp[strlen(csp)-1] == ':')) 
      csp[strlen(csp)-1] = '\0';
    if (swdebug) fprintf(stderr,"SO: %s -> %s %d\n",line,csp,d);
  }
  if (feof(solveonlyfile)) csp[0] = '\0';
}

float readfloatarg(int *xarg, char **argv)
{
  float result;
  
  (*xarg)++;
  if (sscanf(argv[*xarg],"%f",&result) != 1) usage();
  return(result);
}


unsigned int readintarg(int *xsw, int *xarg, char **argv)
{
  unsigned int result;

  if (argv[*xarg][++(*xsw)] == '\0') {
    (*xarg)++;
    if (sscanf(argv[*xarg],"%d",&result) != 1) usage();
  } else {
    if (sscanf(&argv[*xarg][*xsw],"%d",&result) != 1) usage();
  }
  *xsw = strlen(argv[*xarg])-1;
  return(result);
}

int heuristic(char *s, int i, int *oldsplit)
{
  int error=0,newsplit=0; 
  /* printf("heuristic %c, old: %i\n",s[i],*oldsplit); */
  switch (s[i])
    {
    case '1': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=1;     
      newsplit=1;
      break;
    case '2': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=1;     
      newsplit=1;
      break;
    case '3': 
      swfixed = 1;
      swenvcons = 0; 
      swsplitinfo=1;    
      newsplit=1;
       break;
    case '4': 
      swfixed = 1;
      swenvcons = 1; 
      swsplitinfo=1;     
      newsplit=1;
      break;
    case 'A': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=1;
      newsplit=2;
       break;
    case 'B': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=1;
      newsplit=2;
      break;
    case 'C': 
      swfixed = 1;
      swenvcons = 0;
      swsplitinfo=1;
      newsplit=2;
      break;
    case 'D': 
      swfixed = 1;
      swenvcons = 1;
      swsplitinfo=1;
      newsplit=2;
      break; 
    case 'a': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=0;
      newsplit=0;
       break;
    case 'b': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=0;
      newsplit=0;
      break;
    case 'c': 
      swfixed = 1;
      swenvcons = 0;
      swsplitinfo=0;
      newsplit=0;
      break;
    case 'd': 
      swfixed = 1;
      swenvcons = 1;
      swsplitinfo=0;
      newsplit=0;
      break;
    case 's': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=1;     
      newsplit=3;
      break;
    case 't': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=1;     
      newsplit=3;
      break;
    case 'u': 
      swfixed = 1;
      swenvcons = 0; 
      swsplitinfo=1;    
      newsplit=3;
       break;
    case 'v': 
      swfixed = 1;
      swenvcons = 1; 
      swsplitinfo=1;     
      newsplit=3;
      break;
    case 'S': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=1;     
      newsplit=4;
      break;
    case 'T': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=1;     
      newsplit=4;
      break;
    case 'U': 
      swfixed = 1;
      swenvcons = 0; 
      swsplitinfo=1;    
      newsplit=4;
       break;
    case 'V': 
      swfixed = 1;
      swenvcons = 1; 
      swsplitinfo=1;     
      newsplit=4;
      break;
    case 'i': 
      swfixed = 0;
      swenvcons = 0;
      swsplitinfo=1;     
      newsplit=5;
      break;
    case 'j': 
      swfixed = 0;
      swenvcons = 1;
      swsplitinfo=1;     
      newsplit=5;
      break;
    case 'k': 
      swfixed = 1;
      swenvcons = 0; 
      swsplitinfo=1;    
      newsplit=5;
       break;
    case 'l': 
      swfixed = 1;
      swenvcons = 1; 
      swsplitinfo=1;     
      newsplit=5;
      break;
    default: error=1;
    }
  /* printf("swfixed: %i,swenvcons %i, splitfilename: %s, error: %i, i: %i, old: %i, new: %i\n",swfixed,swenvcons,splitfilename,error,i,*oldsplit,newsplit); */
  if ((!error) && (newsplit!=0) && (*oldsplit!=newsplit) ) {
    copysplitfile(newsplit);
    *oldsplit=newsplit;
  }
 
  /* printf("old: %i,  sfn: %s\n",*oldsplit,splitfilename);*/
  return error;
}

void init_headers()
{
  int i;
  char emptystring[] = "empty";
  for(i=0;i<MAXHEADER;i++) {
    headers[i].next=NULL;
    strcpy(headers[i].name,emptystring);
  }
printf("init_headers done\n");
}

int add_headers(FILE *headerfile)
{
  int number=0;
  char line[100],head[100],size[100];
  hp field,new;
  char emptystring[] = "empty";

  while (!feof(headerfile)) {
    while (fgets(line,100,headerfile) && ((line[0] == '\n') || (line[0] == '\0'))); 

    if (sscanf(line,"%s %s",size,head) !=2) {
      fprintf(stderr,"\n*** Header line not found: '%s'\n",line);
      exit(0);
    }

    field=&headers[atoi(size)];

    if(strcmp(field->name,emptystring)==0) {
      strcpy(field->name,head); 
      number++;
      continue;
    }
    
    while((strcmp(field->name,head)!=0) && (field->next!=NULL)) {
      field = field->next;
    }

    if(strcmp(field->name,head)==0) continue;

    if(field->next==NULL) {
      new = malloc(sizeof(hdr));
      strcpy(new->name,head);
      new->next=NULL;
      field->next=new;
      number++;
      continue;
    }
    printf("something is wrong!!\n");
  }
  printf("add_headers done (%i)\n",number);
  return number;
}

int delete_headers(FILE *headerfile)
{
  int number=0;
  char line[100],head[100],size[100];
  hp field,save=NULL;
  char emptystring[] = "empty";


  while (!feof(headerfile)) {
    while (fgets(line,100,headerfile) && ((line[0] == '\n') || (line[0] == '\0'))); 
    if (sscanf(line,"%s %s",size,head) != 2) {
      fprintf(stderr,"\n*** Header line not found: '%s'\n",line);
      exit(0);
    }
    field=&headers[atoi(size)];

    if(strcmp(field->name,emptystring)==0) {
      continue;
    }

    if(strcmp(field->name,head)==0) {
      if(field->next==NULL) {
	strcpy(field->name,emptystring);
	number++;
	continue;
      }
      save=field->next;
      strcpy(field->name,save->name);
      field->next=save->next;
      free(save);
      continue;
    }

    while((strcmp(field->name,head)!=0) && (field->next!=NULL)) {
      save=field;
      field = field->next;
    }

    if(strcmp(field->name,head)==0) {
      if(field->next==NULL) {
	save->next=NULL;
      }
      else {
	save->next=field->next;
      }
      number++;
      free(field);
      continue;
    }
  }
  printf("delete_headers done (%i)\n",number);
  return number;
}

int pop_header(int size, char *head) {
  
  hp field,save=NULL;
  char emptystring[] = "empty";

  field=&headers[size];

  if(strcmp(field->name,emptystring)==0) {
    return 0;
  }

  if(strcmp(field->name,head)==0) {
    if(field->next==NULL) {
      strcpy(field->name,emptystring);
      return 1;
      }
    save=field->next;
    strcpy(field->name,save->name);
    field->next=save->next;
    free(save);
    return 1;
  }

  while((strcmp(field->name,head)!=0) && (field->next!=NULL)) {
    save=field;
    field = field->next;
  }

  if(strcmp(field->name,head)==0) {
    if(field->next==NULL) {
      save->next=NULL;
    }
    else {
      save->next=field->next;
    }
    free(field);
    return 1;
  }
  return 0;
}


int main(int argc, char **argv)
{
  int xarg, xsw;
  int result,i,minnodes,totalnodes,orthcons,oldsplit=0;
  char nextcsp[100];
  char order[30];
  int ordernode[30];
  char first='0';
  double clocks_per_second = (double) CLOCKS_PER_SEC;
  int newsize;
  float newdegree, newlabel, newdensity, factor=1.0;
 

  swhard = 0;
  swdebug = 0;
  swverbose = 0;
  swquiet = 0;
  swsummary = 0;
  swpositive = 0;
  swnegative = 0;
  swposneg = 0;
  swsplitinfo = 0;
  swrandomorder = 0;
  swwqueue = 0;
  swenvcons = 0;
  swfixed = 0;
  swsolveonly = 0;
  swonlypathcons =0;
  swpcprint=0;
  swtesting=0; 
  summaryfile=stdout;
  swmaxvisit = NV_UPPER_LIMIT;
  swfactor=0;
  sworthogonal = 0;
  swallorthogonal = 0;
  swheuristicout = 0;
  swheader=0;
  swminsize=0;
  strcpy(splitfilename,"none");
  
  xarg = 1;
  while ((xarg < argc) && (argv[xarg][0] == '-')) {
    xsw = 1;
    while (argv[xarg][xsw] != '\0') {
      switch (argv[xarg][xsw]) {
      case 'b':
	swbrief = 1;
	swsummary = 1;
	break;
      case 'c':
	swsplitinfo = 1;
	if (argv[xarg][++xsw] == '\0') {
	  if(xarg+1 >= argc) usage();
	  strcpy(splitfilename, argv[++xarg]);
	} else {
	  strcpy(splitfilename, &argv[xarg][xsw]);
	}
	xsw = strlen(argv[xarg])-1;
	readsplitfile(splitfilename);
	break;
      case 'd': 
	swdebug = 1;
	break;
      case 'e':
	swenvcons = 1;
	break;
      case 'f':
	swfixed = 1;
	break;
      case 'g':
	if(xarg+1 >= argc) usage();
	minsize = readintarg(&xsw,&xarg,argv);
	swminsize = 1;
	break;
      case 'h': 
	if(argv[xarg][++xsw]=='a')	 
	  swallorthogonal = 1;
	if(argv[xarg][xsw]=='A') {	 
	  swallorthogonal = 1;
	  swheuristicout = 1;
	  strcpy(heuristicfilename, argv[++xarg]);
	  if ((heuristicfile = fopen(heuristicfilename,"a")) == NULL) {
	    fprintf(stderr,"***Cannot open file '%s'\n",heuristicfilename);
	    exit(-1);
	  }
	}
	if(xarg+1 >= argc) usage();
	strcpy(order, argv[++xarg]);
	sworthogonal = 1;
	swfixed = 1;
	swsplitinfo = 1;
	standardsplitfile();
	xsw = strlen(argv[xarg])-1;
	break;
      case 'l':
	if(argv[xarg][++xsw]=='i') { 
	  swheader=1;
	  if(xarg+1 >= argc) usage();
	  strcpy(headerfilename, argv[++xarg]);
	  if ((headerfile = fopen(headerfilename,"r")) == NULL) {
	    fprintf(stderr,"***Cannot open file '%s'\n",headerfilename);
	    exit(-1);
	  }
	  init_headers();
	  add_headers(headerfile);
	}
	else if(argv[xarg][xsw]=='e') {
	  if(swheader!=1) {printf("Specify option '-li' first.\n");exit(0);} 
	  if(xarg+1 >= argc) usage();
	  strcpy(headerfilename, argv[++xarg]); 
	  if ((headerfile = fopen(headerfilename,"r")) == NULL) {
	    fprintf(stderr,"***Cannot open file '%s'\n",headerfilename);
	    exit(-1);
	  }
	  delete_headers(headerfile);
	}
	xsw = strlen(argv[xarg])-1;
	break;
      case 'm':
	if(argv[xarg][2]=='f') {
	  swfactor=1;
	}
	if(xarg+1 >= argc) usage(); 
	if(swfactor) factor=readfloatarg(&xarg,argv);
	else swmaxvisit= readintarg(&xsw,&xarg,argv);
	maxnodes=swmaxvisit;
	xsw = strlen(argv[xarg])-1;
	break;
      case 'n':
	swpositive = 0;
	swnegative = 1;
	swsummary = 1;
	swposneg = 1;
	break;
      case 'o':
	swsolveonly = 1;
	if (argv[xarg][++xsw] == '\0') {
	  if(xarg+1 >= argc) usage();
	  strcpy(solveonlyname, argv[++xarg]);
	} else {
	  strcpy(solveonlyname, &argv[xarg][xsw]);
	}
	xsw = strlen(argv[xarg])-1;



	if((xarg+1 <= argc) && ( !strcmp(argv[xarg+1],"-1") || !strcmp(argv[xarg+1],"0") || !strcmp(argv[xarg+1],"1"))){
	  taggedwith=readintarg(&xsw,&xarg,argv);
	}
	break;
      case 'p':
	swpositive = 1;
	swnegative = 0;
	swsummary = 1;
	swposneg = 1;
	break;
      case 'q': 
	swquiet = 1;
	break;
      case 'r':
	swrandomorder = 1;
	break;
      case 's':
	swsummary = 1;
	break;
      case 't':
	swtesting=1;
	swsummary=1;
	if (argv[xarg][++xsw] == '\0') {
	  if(xarg+1 >= argc) usage();
	  strcpy(summaryname, argv[++xarg]);
	} else {
	  strcpy(summaryname, &argv[xarg][xsw]);
	}
	xsw = strlen(argv[xarg])-1;
	if ((summaryfile = fopen(summaryname,"w+")) == NULL) {
	  fprintf(stderr,"***Cannot open summary file '%s'\n",summaryname);
	  exit(-1);
	}
	break;
      case 'v':
	swverbose = 1;
	break;
      case 'w':
        if(argv[xarg][2]=='v')	 
	  swwqueue = 2;	
	else swwqueue = 1;
	break;
      case 'x':
        swhard = 1;
        if (argv[xarg][++xsw] == '\0') {
	  if(xarg+1 >= argc) usage();
          strcpy(hardfilename, argv[++xarg]);
        } else {
          strcpy(hardfilename, &argv[xarg][xsw]);
        }
        xsw = strlen(argv[xarg])-1;
        if ((hardfile = fopen(hardfilename,"a")) == NULL) {
          fprintf(stderr,"***Cannot open hard file '%s'\n",hardfilename);
          exit(-1);
        }
        break;
      case 'z': 
	if(argv[xarg][2]=='p')	 
	  swpcprint = 1;	
	swonlypathcons =1;
	break;
      default:
	usage();
      }
      xsw++;
    }
    xarg++;
  }

  if (xarg >= argc) infile = stdin;
  else infile = NULL;

  if (swdebug) {
    putc('I',stderr);
    fflush(stderr);
  }
  initfop();
  if (swdebug) {
    putc('!',stderr);
    fflush(stderr);
  }

  init_statistics();
  size = -1;

  getrusage(RUSAGE_SELF,&rus_glb_before);

  do {
    if (infile == NULL) {
      strcpy(infilename,argv[xarg]);
      if ((infile = fopen(infilename,"r")) == NULL) {
	fprintf(stderr,"\n*** Cannot open '%s' for reading\n",infilename);
	exit(-1);
      }
    }

    while (!feof(infile)) {
      if (swsolveonly) getnextcsp(nextcsp);
      if (swdebug) fprintf(stderr,"Next: %s\n",nextcsp);

      do {
#if defined(DYNAMIC)
	result = parsecsp_vsd(infile, MAXCSP-1, prevmax, &maxnodeid,csptype,&csp);
	
	if (swdebug) fprintf(stderr,"Check: %s\n",csptype);

	if (size == -1) { /* First CSP */
	  if (swtesting) {
	    parse_type(csptype,&newsize,&newlabel,&newdegree,&newdensity);
	    strcpy(lasttype,csptype);
	  }
	  prevmax = maxnodeid + 1;
	  size = prevmax;
	  mark = (char **)malloc(prevmax*sizeof(char *));
	  if((mark[0] = (char *)malloc(prevmax*prevmax*sizeof(char))) == NULL){
	    fprintf(stderr,"\n*** No more memory available\n");
	    exit(-1);
	  }
	  for(i=1;i<prevmax;i++)
	    mark[i] = mark[0] + i * prevmax;
	  if(swwqueue){
	    entry = (struct entry **)malloc(MAXWT*sizeof(struct entry *));
	    if((entry[0] = (struct entry *)malloc(MAXWT*prevmax*prevmax*sizeof(struct entry)))== NULL){
	      fprintf(stderr,"\n*** No more memory available\n");
	      exit(-1);
	    }
	    for(i=1;i<MAXWT;i++)
	      entry[i] = entry[0] + i * prevmax * prevmax;
	  }
	  if(swfixed || sworthogonal) {
	    consval = (int **)malloc(prevmax*sizeof(int *));
	    if((consval[0] = (int *)malloc(prevmax*prevmax*sizeof(int))) ==NULL){
	      fprintf(stderr,"\n*** No more memory available\n");
	      exit(-1);
	    }
	    for(i=1;i<prevmax;i++)
	      consval[i] = consval[0] + i * prevmax;
	  }
	}


	if (size != prevmax) size = 0;
	if (result == -1) exit(-1);
	if (prevmax <= maxnodeid) { /* new CSP larger than previous ones */
	  prevmax = maxnodeid+1;

	  free((void *) mark[0]); free((void *) mark); 

	  mark = (char **)malloc(prevmax*sizeof(char *));
	  if((mark[0] = (char *)malloc(prevmax*prevmax*sizeof(char))) ==NULL){
	    fprintf(stderr,"\n*** No more memory available\n");
	    exit(-1);
	  }
	  for(i=1;i<prevmax;i++)
	    mark[i] = mark[0] + i * prevmax;

	  if(swwqueue) {

	    free((void *) entry[0]); free((void *) entry); 

	    entry = (struct entry **)malloc(MAXWT*sizeof(struct entry *));
	    if((entry[0] = (struct entry *)malloc(MAXWT*prevmax*prevmax*sizeof(struct entry)))==NULL){
	      fprintf(stderr,"\n*** No more memory available\n");
	      exit(-1);
	    }
	    for(i=1;i<MAXWT;i++)
	      entry[i] = entry[0] + i * prevmax * prevmax;
	  }
	  if(swfixed || sworthogonal) {

	    free((void *) consval[0]); free((void *) consval); 

	    consval = (int **)malloc(prevmax*sizeof(int *));
	    if((consval[0] = (int *)malloc(prevmax*prevmax*sizeof(int)))==NULL){
	      fprintf(stderr,"\n*** No more memory available\n");
	      exit(-1);
	    }
	    for(i=1;i<prevmax;i++)
	      consval[i] = consval[0] + i * prevmax;
	  }
	}

#else
	result = parsecsp(infile, MAXCSP-1,&maxnodeid,csptype,csp);
	
	if (swdebug) fprintf(stderr,"Check: %s\n",csptype);
	
	if (size == -1) {
	  size = maxnodeid + 1;
	  if(swtesting) {
	    parse_type(csptype,&newsize,&newlabel,&newdegree,&newdensity);     
	  }
	}
	if (size != maxnodeid + 1) size = 0;
	if (result == -1) exit(-1);
#endif

      } while ((swsolveonly) &&
	       (strcmp(nextcsp,csptype) != 0) &&
	       (result != 0));

      if (result == 0) break;
      
      if ((trial >= MAXTRIAL) && (swsummary)) {
	fprintf(stderr,"\n*** More than %d CSPs given! Quitting now\n",MAXTRIAL);
	getrusage(RUSAGE_SELF,&rus_glb_after);
	cumcputicks = (double)(((double) rus_glb_after.ru_utime.tv_usec+clocks_per_second*(double)rus_glb_after.ru_utime.tv_sec) - ((double)rus_glb_before.ru_utime.tv_usec+clocks_per_second*(double)rus_glb_before.ru_utime.tv_sec))/clocks_per_second;  
	print_summary(swbrief,csptype,summaryfile);
	if (swquiet) exit(!consistent);
	else exit(0);
      }

      pcops = 0;
      pcits = 0;
      nodesvisited = 0;
      searchdepth = 0;
      minnodes = maxnodes;
      maxdepth = 0;
      pathconsistent = 0;
      if (swdebug) {
	fprintf(stderr,"\nSolving %s:\n", csptype);
	fflush(stderr);
      }   

      if(swheader) {
	if(pop_header(maxnodeid,csptype)!=1) continue;
      }

      if(swminsize) {
	if(maxnodeid+1<minsize) continue; 
      }

      getrusage(RUSAGE_SELF,&rus_lcl_before);
      
      if(swfactor){ 
	swmaxvisit=factor*(maxnodeid+1);
	maxnodes=swmaxvisit;
      }

      if(sworthogonal){
	/* use different heuristics */
	i=0;
	totalnodes=0;
	minnodes = maxnodes;
	first = '0';
	consistent = -1;
	while (order[i] != '\0') {
	  nodesvisited=0;
	  searchdepth=0;
	  ordernode[i]=0;
	  if(heuristic(order,i,&oldsplit)) {
	    fprintf(stderr, "**** `%c' is not a valid method! \n Please choose another <order> argument for the `-h[a|A <file>] <order>' switch.\n",order[i]);
	    exit(-1);
	  }

	  orthcons = global_cons(csp,maxnodeid,-1,-1); 

	  if(orthcons!=-1) ordernode[i]=nodesvisited;
	  if(nodesvisited < minnodes) {
	    minnodes=nodesvisited;
	    if(!swheuristicout) swmaxvisit=minnodes;
	    consistent = orthcons;
	    first=order[i];
	  }
	  totalnodes+=nodesvisited;
	
	  if((!swallorthogonal) && (consistent != -1)) break;
	  i++;
	}
	if(swheuristicout) heuristic_summary(minnodes, order, first, ordernode, heuristicfile);
        swmaxvisit=maxnodes;
      }
      else if(swonlypathcons){
	/* check only path-consistency */
	pathconsistent=path_cons(csp,maxnodeid, -1,-1);
	if(swpcprint && pathconsistent) printcsp(stdout,maxnodeid,csptype,csp);
	consistent=pathconsistent;
      }
      else{
	/* check global consistency */
	consistent = global_cons(csp,maxnodeid,-1,-1);
      }
      getrusage(RUSAGE_SELF,&rus_lcl_after);

      if(swtesting && parse_type(csptype,&newsize,&newlabel,&newdegree,&newdensity)) {
       print_summary(swbrief,lasttype,summaryfile);
       strcpy(lasttype,csptype);
      }
     

      if (swdebug) {
	putc('\n',stderr);
	if (swquiet){ 
	  fprintf(stderr,"... solved. path-cons=%d, cons=%d\n",
		  pathconsistent,consistent);
	}
	fflush(stderr);
      }

      /* statistics */
     
      cputicks = (double)(((double) rus_lcl_after.ru_utime.tv_usec+clocks_per_second*(double)rus_lcl_after.ru_utime.tv_sec) - ((double)rus_lcl_before.ru_utime.tv_usec+clocks_per_second*(double)rus_lcl_before.ru_utime.tv_sec))/clocks_per_second;  
     
      if(consistent != -1) {
	if (consistent) cntconsistent++;
	if (pathconsistent) cntpathconsistent++;
	cntsize += maxnodeid+1;
	if (!swposneg || 
	    ((consistent) && swpositive) || 
	    ((!consistent) && swnegative)) {
	  tpcops[rectrial] = pcops;
	  tpcits[rectrial] = pcits;
	  tnodesvisited[rectrial] = nodesvisited;
	  tmaxdepth[rectrial] = maxdepth;
	  tcputicks[rectrial] = cputicks; 
	  tconsistent[rectrial] = consistent;
	  rectrial++;
	}
	trial++;
      }
   /*    if ((swhard) && (nodesvisited >= HARD_NODES)) printcsp(hardfile,maxnodeid,csptype,csp); */
      if ((swhard) && (consistent == -1)) printcsp(hardfile,maxnodeid,csptype,csp);

      /* output result */
      if (!swquiet) {
	if (!swverbose) {
	  if (!sworthogonal)
	    fprintf(stderr,"%s: %d\n", csptype, consistent);
	  else 
	    fprintf(stdout,"%s: %d, nodes: %i, method: %c\n", csptype, consistent,minnodes,first);
	}
	else
	  fprintf(stdout,"%s: %d  CPU=%f, PC-ops=%d, PC-it=%d, Nodes=%d,Depth=%d\n",
		 csptype, consistent,
		 cputicks, pcops, pcits, nodesvisited, maxdepth);
      }

      
    }
    xarg++;
    if (infile != stdin) fclose(infile);
    infile = NULL;
  } while (xarg < argc);

  getrusage(RUSAGE_SELF,&rus_glb_after); 
  cumcputicks = (double)(((double) rus_glb_after.ru_utime.tv_usec+clocks_per_second*(double)rus_glb_after.ru_utime.tv_sec) - ((double)rus_glb_before.ru_utime.tv_usec+clocks_per_second*(double)rus_glb_before.ru_utime.tv_sec))/clocks_per_second;  
 
  if (swsummary) print_summary(swbrief,csptype,summaryfile);
  if (swquiet) exit(!consistent);
  else exit(0);
}    

